#!/corp/global/tools/bin/python3.4
import re
import os
import sys
import pexpect
from pexpect import pxssh
import pprint
import getpass
import argparse
import unittest
import pycurl
from time import strftime
from functools import reduce
from io import BytesIO
from subprocess import Popen, PIPE, call
import xml.etree.ElementTree as ET
import itertools

url      = ''
proto    = ''
RESTCONF = 1

responses = {#Based on RFC 2616.
    100: ('Continue', 'Request received, please continue'),
    101: ('Switching Protocols',
          'Switching to new protocol; obey Upgrade header'),

    200: ('OK', 'Request fulfilled, document follows'),
    201: ('Created', 'Document created, URL follows'),
    202: ('Accepted',
          'Request accepted, processing continues off-line'),
    203: ('Non-Authoritative Information', 'Request fulfilled from cache'),
    204: ('No Content', 'Request fulfilled, nothing follows'),
    205: ('Reset Content', 'Clear input form for further input.'),
    206: ('Partial Content', 'Partial content follows.'),

    300: ('Multiple Choices',
          'Object has several resources -- see URI list'),
    301: ('Moved Permanently', 'Object moved permanently -- see URI list'),
    302: ('Found', 'Object moved temporarily -- see URI list'),
    303: ('See Other', 'Object moved -- see Method and URL list'),
    304: ('Not Modified',
          'Document has not changed since given time'),
    305: ('Use Proxy',
          'You must use proxy specified in Location to access this '
          'resource.'),
    307: ('Temporary Redirect',
          'Object moved temporarily -- see URI list'),

    400: ('Bad Request',
          'Bad request syntax or unsupported method'),
    401: ('Unauthorized',
          'No permission -- see authorization schemes'),
    402: ('Payment Required',
          'No payment -- see charging schemes'),
    403: ('Forbidden',
          'Request forbidden -- authorization will not help'),
    404: ('Not Found', 'Nothing matches the given URI'),
    405: ('Method Not Allowed',
          'Specified method is invalid for this server.'),
    406: ('Not Acceptable', 'URI not available in preferred format.'),
    407: ('Proxy Authentication Required', 'You must authenticate with '
          'this proxy before proceeding.'),
    408: ('Request Timeout', 'Request timed out; try again later.'),
    409: ('Conflict', 'Request conflict.'),
    410: ('Gone',
          'URI no longer exists and has been permanently removed.'),
    411: ('Length Required', 'Client must specify Content-Length.'),
    412: ('Precondition Failed', 'Precondition in headers is false.'),
    413: ('Request Entity Too Large', 'Entity is too large.'),
    414: ('Request-URI Too Long', 'URI is too long.'),
    415: ('Unsupported Media Type', 'Entity body in unsupported format.'),
    416: ('Requested Range Not Satisfiable',
          'Cannot satisfy request range.'),
    417: ('Expectation Failed',
          'Expect condition could not be satisfied.'),

    500: ('Internal Server Error', 'Server got itself in trouble'),
    501: ('Not Implemented',
          'Server does not support this operation'),
    502: ('Bad Gateway', 'Invalid responses from another server/proxy.'),
    503: ('Service Unavailable',
          'The server cannot process the request due to a high load'),
    504: ('Gateway Timeout',
          'The gateway server did not receive a timely response'),
    505: ('HTTP Version Not Supported', 'Cannot fulfill request.'),
    }

header = ["Resource-Depth: 20", "Accept: application/vnd.operational-state.resource+xml"]
conf_params = ['DEVICEIP', 'DEVICEUNAME', 'DEVICEPASSWD', 'RESTUNAME', 'RESTPASSWD', 'OUTPUTDIR', 'CONFIGFILE']
rc_conf_params = ['DEVICEIP', 'DEVICEUNAME', 'DEVICEPASSWD', 'RESTCONFUNAME', 'RESTCONFPASSWD', 'OUTPUTDIR', 'CONFIGFILE']
param_dict = {}
keys_dict = {}
grouped_keys = []

sub_tree_node = None
gen_rest_dump = False
uri_list = []
test_result = 'FAIL'
ret_code = ''
outfile =''
base_path =  os.getcwd()#os.path.dirname(os.path.realpath(__file__))

def create_rest_tree():
    print("\nGenerating flattened YANG tree from the operational YANG FILES")
    print("*********************************************************************************************\n")
    os.system(" "+base_path+"/../../../../../../../target26/confd/LATEST/x86_64/bin/pyang -o rest_tree -f tree -p yang/:"+base_path+"/../../../../../src/yang "+base_path+"/../../../../../src/yang/*oper*.yang >> log_file 2>&1")
    print("*********************************************************************************************\n")

def doSsh(user, password, sport, host, config_file):
    child = pexpect.spawn('ssh '+user+'@'+host)
    child.expect(".*assword:")
    child.send(password+"\r")
    #child.expect('.*to proceed.')
    #child.send('\003')
    child.expect('>')
    child.sendline('noscli')
    child.expect('#')
    child.sendline('conf t\r\r\r\r')
    child.expect('#')
    child.sendline('\r'.join([line.strip() for line in open(config_file, 'r')])+'\r')

def optional_args():
    global dest_dir
    global sub_tree_node
    global dump_gen
    global url
    global proto

    parser = argparse.ArgumentParser(description='Run Operational commands test suite')

    parser.add_argument('-r', dest="rest_conf", action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-t', dest="sub_tree_node", type=str, action='store',
                   help='Subtree to start the suite')
    parser.add_argument('-g', dest="dump_gen", action='store_true',
                   help='Generate the complete Operational URI dump')

    args = parser.parse_args()

    if args.rest_conf:
        proto = RESTCONF
        url = "https://"
        help_string = 'Generate the complete RESTCONF Operational dump'
    else:
        url = "http://"#10.24.12.85"

    sub_tree_node = args.sub_tree_node or None
    dump_gen = args.dump_gen

    if args.rest_conf:
        proto = RESTCONF
        url = "https://"
    else:
        url = "http://"#10.24.12.85"

    if sub_tree_node and dump_gen:
        print("Error: Options '-g' and '-t' cannot coexist")
        exit()

    print("\n*********************************************************************************************\n")
    if sub_tree_node is not None:
        print("Starting from sub tree "+sub_tree_node)

def read_conf_file():
    fname = ''
    global conf_params
    global rc_conf_params
    global param_dict

    if proto == RESTCONF:
        fname = "rcBreezerc"
        params = rc_conf_params
    else:
        fname = "rBreezerc"
        params = conf_params

    if os.path.exists(os.getcwd()+"/"+fname):
        for line in open(fname, "r"):
            if len(line.strip()):
                if line.startswith("#"):
                    pass
                else:
                    [key, val] = line.strip().split("=")
                    param_dict[key] = val
        #param_dict = dict([line.strip().split("=") for line in open("rBreezerc", "r")])
    else:
        conf = open(fname, "w")
        for param in params:
            val = input(param+"[Empty]=")
            conf.write(param+"="+val+"\n")
            param_dict[param] = val
        conf.close()

class read_tree:
    global uri_list
    global keys_dict

    def _init_(self):
        print("Inited Class")

    def get_level(self, len):
        return (len-8)//3

    def print_uri(self):
        cnt = 0
        pat = r"(.*\+--ro) ([A-z-\d]+)([\?\*]?)\s*(\[(.*)\])?"
        sub_tree = ""
        keys = []
        is_leaf = 1
        parent_node = ""
        mod_name = ""

        if proto == RESTCONF:
            list_elem = ["/restconf/data"]
        else:
            list_elem = ["/rest/operational-state"]

        for line in open("rest_tree", "r"):
            if proto == RESTCONF:
                if "module:" in line:
                    mod_name = line.split(":")[1].split("-operational")[0].strip()

            if "+" in line:
                match = re.match(pat, line)
                final_uri = ""
                level = self.get_level(len(match.group(1)))+1
                node = match.group(2)

                if match.group(2) in keys:
                    keys_dict[match.group(2)] = []
                    keys.remove(match.group(2))
                    continue

                if level >= len(list_elem):
                    list_elem.append(node)
                else:
                    list_elem[level] = node
                    list_elem = list_elem[:level+1]

                if not match.group(3):
                    is_leaf = 0
                    parent_node = "/".join(list_elem)
                    sub_tree = parent_node
                    if match.group(5):
                        keys = match.group(5).split()
                        sub_tree += "/"+node

                final_uri  = "/".join(list_elem)

                if proto == RESTCONF:
                    final_uri = final_uri[:15]+mod_name+":"+final_uri[15:]

                if not is_leaf:
                    if len(keys):
                        sub_tree += "/"+"/".join(keys)
                        list_elem[level] = node+"/"+'/'.join(['val_'+x+'_val' for x in keys])
                    is_leaf = 1

                uri_list.append(final_uri)

def set_test_result(code):
    global ret_code
    global test_result
    ret_code = str(code)+" "+responses[code][0]
    if 200 == code:
        test_result = "PASS"
    else:
        test_result = "FAIL"


def gen_valid_combos():
    valid_combos = []
    val_list = []
    new_val_list = []
    cont = 0

    for ele in grouped_keys:
        if ele[0] in keys_dict.keys():
            val_list.append([keys_dict[x] for x in ele])
            cont = 1
    if cont:
        for ele in val_list:
            new_val_list.append(list(map(list, zip(*ele))))
        combos = list(map(list, list(zip(*new_val_list))))
        for ele in combos:
            valid_combos.append(list(itertools.chain.from_iterable(ele)))
    return valid_combos

def perform_query(cmd):
    rest_req = pycurl.Curl()
    rest_req.setopt(pycurl.CONNECTTIMEOUT, 30)
    if proto == RESTCONF:
        rest_req.setopt(rest_req.FOLLOWLOCATION, True)
        rest_req.setopt(pycurl.SSL_VERIFYPEER, 0)
        rest_req.setopt(pycurl.SSL_VERIFYHOST, 0)
    else:
        rest_req.setopt(pycurl.HTTPHEADER, header)
    if proto == RESTCONF:
        rest_req.setopt(pycurl.USERPWD, param_dict["RESTCONFUNAME"] + ':' + param_dict["RESTCONFPASSWD"])
    else:
        rest_req.setopt(pycurl.USERPWD, param_dict["RESTUNAME"] + ':' + param_dict["RESTPASSWD"])
    rest_req.setopt(pycurl.HTTPGET, 1)
    print("Executing GET request for ", cmd)
    buffer = BytesIO()
    rest_req.setopt(rest_req.WRITEDATA, buffer)
    rest_req.setopt(pycurl.URL, url + cmd)
    rest_req.perform()
    output = buffer.getvalue().decode('utf-8')
    set_test_result(rest_req.getinfo(pycurl.HTTP_CODE))
    format_output(cmd, output, rest_req.errstr())
    buffer.flush()
    rest_req.close()

def walk_node(cmd):
    global param_dict
    req = ""

    if "val_" in cmd:
        for ele in gen_valid_combos():
            req = ""
            i = 0
            for sub_ele in cmd.split("val"):
                if sub_ele.startswith("_"):
                    req +=str(ele[i])
                    i += 1
                else:
                    req += sub_ele
            perform_query(req)
    else:
        perform_query(cmd)

def write_test_header(cmd, err_str):
    outfile.write("*********************************************************************************************************************************\n")
    outfile.write("================                           =============\n")
    outfile.write("|      TEST    |                           |    "+test_result+"    |\n")
    outfile.write("================                           =============\n")
    outfile.write("*********************************************************************************************************************************\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")
    outfile.write("URI".ljust(10)+": "+cmd+"\n")
    outfile.write("STATUS".ljust(10)+": "+ret_code+"\n")
    outfile.write("ERRORS".ljust(10)+": "+err_str+"\n")
    outfile.write("-----------------------------------------------------------------------------------------------\n")

def update_keys(key, val):
    global keys_dict
    if key in keys_dict.keys():
        keys_dict[key].append(val)

def format_output(cmd, buffer, err_str):
    global keys_dict
    node = cmd.rsplit("/", 1)[1]
    pat = '<'+node+'.*".*">\n?(.*)\n?</'+node+'>'
    ns = '{urn:brocade.com'
    if len(buffer):
        write_test_header(cmd, err_str)
        outfile.write("Output Dump:\n")
        outfile.write("-----------------------------------------------------------------------------------------------\n")
        outfile.write(buffer)
        outfile.write("-----------------------------------------------------------------------------------------------\n")
        outfile.write("Formatted Output:\n")
        outfile.write("-----------------------------------------------------------------------------------------------\n")
        try:
            parser = ET.XMLParser(encoding="utf-8")
            tree = ET.fromstring("<root>\n"+buffer+"</root>\n", parser=parser)
        except ET.ParseError as e:
            print("Parsing of the response failed with the following reason\n"+str(e)+"\n")
            print("Dumping the Buffer:\n"+buffer)

        for child in tree:
            if not child.getchildren():
#No clue as to why this check is required. Sometimes the namespace is not present in the response.
                if child.tag.find(ns) > 0:
                    tag = child.tag.split("}")[1].strip()
                else:
                    tag = child.tag
                text = child.text
                if not text:
                    text = ""
                outfile.write(tag.ljust(50)+' = '+text+"\n")
                update_keys(tag, child.text)
            else:
                for sub_child in child.getchildren():
                    if len(sub_child.tag.split("}")) > 1:
                        tag = sub_child.tag.split("}")[1].strip()
                        text = sub_child.text
                        if not text:
                            text = ""
                        outfile.write(tag.ljust(50)+' = '+text+"\n")
                        update_keys(tag, sub_child.text)

        outfile.write("-----------------------------------------------------------------------------------------------\n")
        outfile.write("*********************************************************************************************************************************\n\n\n")
    elif test_result == "PASS" or len(ret_code):
        write_test_header(cmd, err_str)
    else:
        print("EXECUTION OF TEST CASES FAILED. IT MIGHT BE DUE TO THE FOLLOWING REASONS.\
               \n1.URI NOT FOUND. \
               \n2. DEVICE CRASHED.\
               \n3. CONNECTIVITY LOST.")

def group_keys(uri):
    #Written to track all the keys that are together and not perform permutations
    #over them.
    global grouped_keys
    grouped_keys = []
    old_group = 2
    new_group = 1
    copy = 0
    for ele in uri.split("_"):
        if ele == "val/val":
            copy = old_group
        elif ele.endswith("val"):
            copy = new_group
        else:
            if copy == new_group:
                grouped_keys.append([ele])
            elif copy == old_group:
                grouped_keys[-1].append(ele)
            copy = 0

def main():
    global sub_tree_node
    global param_dict
    global outfile
    global url
    dump_fname    = ''
    sub_tree_list = []

    optional_args()
    read_conf_file()
    url += param_dict["DEVICEIP"]
    if not os.path.exists("rest_tree"):
        create_rest_tree()

    if os.path.exists(param_dict["CONFIGFILE"]):
        print("Please wait...Pushing Config to the device")
        doSsh(param_dict["DEVICEUNAME"], param_dict["DEVICEPASSWD"], 22, param_dict["DEVICEIP"], param_dict["CONFIGFILE"])
    
    outputdir = param_dict["OUTPUTDIR"]
    if ("" != outputdir):
       outputdir += "/"
    if proto == RESTCONF:
        #root_uri = "/rest/operational-state/"
        root_uri = "/restconf/data/"
        dump_fname = "REST_CONF_OPER_URI_DUMP"
    else:
        #root_uri = "/restconf/data/"
        root_uri = "/rest/operational-state/"
        dump_fname = "REST_OPER_URI_DUMP"

    file_name = "testOutput_"+strftime("%Y-%m-%d_%H-%M-%S")
    uri_dump = read_tree()
    uri_dump.print_uri()

    if not os.path.exists(outputdir+dump_fname):
        out = open(outputdir+dump_fname , "w")
        for ele in uri_list:
            out.write(ele+"\n")
        out.close()
    if dump_gen:
        print("For URI dump refer to ", outputdir+dump_fname)
        print("\n*********************************************************************************************\n")
        exit()

    if sub_tree_node is None:
        sub_tree_node=root_uri
    outfile = open(outputdir+file_name, "w")
    print("Using "+outfile.name+" for writing output")
    print("Executing GET request for ",sub_tree_node)

    group_keys(sub_tree_node)
    walk_node(sub_tree_node)

    for ele in uri_list:
        if ele.startswith(sub_tree_node+"/"):
            group_keys(ele)
            walk_node(ele)

    outfile.close()
    print("*********************************************************************************************\n")
    print("For URI dump refer to ", outputdir+dump_fname)
    print("Output is written to ", outputdir+file_name)
    print("*********************************************************************************************\n")

if __name__ == '__main__':
    main()
